{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE NamedFieldPuns #-}

module Pages.TechDemo (techDemoPage) where

import qualified Data.ByteString.Lazy as LBS
import qualified Data.Text as T
import qualified Data.Text.IO as T
import qualified Data.Text.Encoding as T
import Data.ByteString.Base32 (encodeBase32Unpadded)
import Control.Monad.IO.Class (liftIO)
import Data.List.NonEmpty (NonEmpty(..))
import Network.HTTP.Req (HttpException, GET(GET), NoReqBody(NoReqBody), https, req, ignoreResponse, responseHeader, runReq, defaultHttpConfig)
import Data.Maybe (fromMaybe)
import Control.Monad (void)
import Control.Exception (try)
import Data.Serialize (encode)

import Tahoe.CHK.Capability (CHK(..), Reader(..), Verifier(..), dangerRealShow, pCapability)
import Tahoe.CHK.Upload (UploadResult(..), memoryUploadableWithConvergence, store)
import Tahoe.CHK.Types (Parameters(..), StorageServer)
import Tahoe.Server (nullStorageServer)

import Obelisk.Frontend
import Reflex.Dom.Core
import Language.Javascript.JSaddle (eval, liftJSM)

import Text.Megaparsec (ParseErrorBundle, parse)

import FrontendPaths (getFilesDir)

techDemoPage :: ObeliskWidget t r m => m (Event t ())
techDemoPage = do
    el "h1" $ text "Tech Demo"

    -- The API for encoding a CHK requires some servers to upload to.  We
    -- don't have any real servers to upload to yet but we can make up this
    -- no-op server that's good enough.  This "behavior" value always has a
    -- NonEmpty of StorageServers.  We can pass it around to any API that
    -- needs servers and the value can be sampled whenever it is needed.
    -- This centralizes responsibility for knowing which servers to use
    -- here.
    --
    -- we represent this as a NonEmpty because the current implementation
    -- errors on an empty server list.  Also it is not clear what it really
    -- means to upload to no servers.
    let serversBehavior = constant (nullStorageServer :| [])
        -- It also requires some encoding parameters.  Handle those
        -- likewise.
        paramsBehavior = constant Parameters
                         -- Note this is an unreasonably small segment size.
                         { paramSegmentSize = 128
                         , paramRequiredShares = 7
                         , paramHappyShares = 1
                         , paramTotalShares = 11
                         }

    -- Place an input field into the page and retain a reference to it.
    el "span" $ text "Encode some plaintext: "
    (_, inputWidget) <- el' "span" $ inputElement def

    -- Get an event that triggers any time the contents of the input field
    -- change (a lot of changes!).  It triggers with the utf-8 encoded
    -- representation of the contents.
    let plaintextEv = LBS.fromStrict . T.encodeUtf8 <$> updated (_inputElement_value inputWidget)

    -- Use the behaviors and the plaintext input event to encode the data
    -- and make an event that triggers whenever this is done, with the
    -- "upload result" (quirk of the current tahoe-chk API) as its value.
    uploadedEv <- uploadEv serversBehavior paramsBehavior plaintextEv

    -- Get an event that triggers with the full string representation of the
    -- capability resulting from encoding the plaintext.
    let theCapTextEv = dangerRealShow . CHKReader . uploadResultReadCap <$> uploadedEv

    -- Put the capability string into the page.
    theCapTextDyn <- holdDyn "" theCapTextEv
    el "p" $ dynText theCapTextDyn

    -- Show that we can parse capabilities as well as generate them.
    el "p" parseCapDemo

    -- Do some file I/O and put the results on the page.
    el "p" filesDemoWidget

    -- Some network I/O
    el "p" networkWidget

    -- Demonstrate Obelisk "pre-rendering".
    el "p" prerenderWidget

    -- Add a button to get out of here.
    button "Leave the tech demo"

-- | Accept a capability string as input, parse it, and render something about
-- it.
parseCapDemo :: ObeliskWidget t r m => m ()
parseCapDemo =
    el "div" $ do
      el "span" $ text "Parse a cap: "
      -- Place an input field into the page and retain a reference to it.
      capInputWidget <- inputElement def

      -- `Dynamic t T.Text` holding the value of the text input.
      let capTextDyn = _inputElement_value capInputWidget

      -- `Dynamic t (Either ParseErrorBundle s e) CHK` holding a parse error
      -- or a parsed capability.
      let parsedCap = parse pCapability "" <$> capTextDyn

      dyn_ (renderCap <$> parsedCap)

-- | The rendering part of parseCapDemo.
renderCap :: ObeliskWidget t r m => Either (ParseErrorBundle s e) CHK -> m ()
renderCap (Left _) = text "(unparseable)"
renderCap (Right (CHKVerifier Verifier { storageIndex, fingerprint, required, total, size })) = do
  el "div" $ text ("Storage index: " <> T.toLower (encodeBase32Unpadded storageIndex))
  el "div" $ text ("Fingerprint: " <> T.toLower (encodeBase32Unpadded fingerprint))
  el "div" $ text ("Required shares: " <> T.pack (show required))
  el "div" $ text ("Total shares: " <> T.pack (show total))
  el "div" $ text ("Plaintext size: " <> T.pack (show size))

renderCap (Right (CHKReader r)) = do
  el "div" $ text ("Read key: " <> (T.toLower . encodeBase32Unpadded . encode . readKey) r)
  renderCap (Right . CHKVerifier . verifier $ r)

-- | Given an event of a plaintext create an event of the CHK capability
-- object representing that plaintext.  Note that since the data is not
-- uploaded anywhere the capability cannot actually be used to recover the
-- plaintext.
uploadEv ::
    ObeliskWidget t r m =>
    -- | A behavior of the servers we could upload to.
    Behavior t (NonEmpty StorageServer) ->
    -- | A behavior of the encoding parameters we should use.
    Behavior t Parameters ->
    -- | An event that triggers with plaintext that we should encrypt and
    -- upload.
    Event t LBS.ByteString ->
    -- | An event that triggers with an upload result after an upload
    -- succeeds.
    m (Event t UploadResult)
uploadEv servers parameters plaintextEv = do

  -- Ignore empty strings, the encoder does not support them (XXX
  -- unfortunately it fails by going into an infinite loop).
  let nonEmptyPlaintext = (not . LBS.null) `ffilter` plaintextEv

  -- Attach upload parameters to the plaintext
  let paramsAndPlaintext = attach parameters nonEmptyPlaintext

  -- Whenever the plaintext changes, construct an "uploadable" value from it
  -- that can be used below.
  uploadableEv <-
      performEvent $ ffor paramsAndPlaintext $ \(params, plaintext) ->
          liftIO $
                 memoryUploadableWithConvergence "secret" (fromIntegral $ LBS.length plaintext) plaintext params

  -- Attach the server list to the uploadable value
  let inputsReady = attach servers uploadableEv

  -- Run the actual upload of the uploadable to the current servers and return
  -- the result.
  performEvent $ ffor inputsReady $ \(aServer :| moreServers, uploadable) ->
      liftIO $ store (aServer:moreServers) uploadable


filesDemoWidget :: ObeliskWidget t r m => m ()
filesDemoWidget =
  -- Rendering the main body of the files demo widget on the server and client
  -- (both part of the frontend) produces different results.  This causes a
  -- warning about "hydration failure" indicating reflex-dom has gotten
  -- confused by this difference.  There's no particular reason to render the
  -- files demo widget on the server.  We can use `prerender_ blank` to cause
  -- the server to render `blank` and let the client be the only renderer of
  -- the real content.
  prerender_ blank $ do
    (runButton, _) <- el' "button" $ text "Run File I/O"
    let runEv = domEvent Click runButton

    -- Demonstrate writing and reading data in an execution-context-defined
    -- application-private location.
    ev <- performEvent . ffor runEv . const . liftIO $
          do
            p <- getFilesDir
            case p of
              Nothing -> pure Nothing
              Just p' ->
                  do
                    -- Be sure to do strict IO here, which Data.Text.IO does,
                    -- otherwise the lazy write and the lazy read conflict
                    -- with each other in surprising ways (eg, resulting in
                    -- in "resource is busy" error from the Haskell runtime).
                    -- Another solution would be to use getContents or
                    -- withFile or some such.
                    T.writeFile path "Hello, world!"
                    content <- T.readFile path
                    pure (Just (p', content))
                  where
                    path = p' <> "/" <> "somedata.txt"

    filesDyn <- holdDyn Nothing ev
    el "p" $ dynText (T.pack . show <$> filesDyn)

-- | Demonstrate doing some network I/O in a way that's integrated with
-- Reflex-DOM.
networkWidget :: ObeliskWidget t r m => m ()
networkWidget =
  -- See filesDemoWidget for an explanation of the prerender_ call.
  prerender_ blank $ do
    (runButton, _) <- el' "button" $ text "Run Request"
    let runEv = domEvent Click runButton
    resultEv <- performEvent $ ffor runEv $ \_ -> liftIO getResult
    resultDyn <- holdDyn Nothing resultEv
    el "div" $ do
      el "span" $ text "Resource age:"
      el "span" $ dynText (fromMaybe "<unknown>" <$> resultDyn)
    where
      request = req GET (https "private.storage") NoReqBody ignoreResponse mempty
      getResult :: IO (Maybe T.Text)
      getResult = do
        respE <- try $ runReq defaultHttpConfig request
        case respE of
          Left (err :: HttpException) -> pure . Just . T.pack . ("Error: " <>) . show $ err
          Right resp ->
              pure $ T.decodeUtf8 <$> responseHeader resp "age"


prerenderWidget :: ObeliskWidget t r m => m ()
prerenderWidget =
    -- `prerender` and `prerender_` let you choose a widget to run on the server
    -- during prerendering and a different widget to run on the client with
    -- JavaScript. The following will generate a `blank` widget on the server and
    -- print "Hello, World!" on the client.
    prerender_ blank $ liftJSM $ void $ eval ("console.log('Hello, World!')" :: T.Text)
